unit hiDataAccumulator;

{.$WARNINGS OFF}

interface

uses
  Windows, Kol, Share, Debug;

type

  
  ThiDataAccumulator = class(TDebug)
    private
      FBuffer: string;
      FCurrentLength: Integer;
      FTotalLength: Cardinal;
      
      procedure SetBufferSize(Value: Integer);
    public
      _prop_Delimiter: string;
      _prop_Length: Integer;
      _prop_MaxLength: Integer;
      _prop_Mode: Integer;

            
      _event_onError: THI_Event;
      _event_onFinish: THI_Event;
      _event_onData: THI_Event;
      
      _data_Data: THI_Event;
      _data_Delimiter: THI_Event;
      _data_Length: THI_Event;

      property _prop_BufferSize: Integer write SetBufferSize;

      constructor Create;

      procedure _work_doFinish(var _Data: TData; Index: Word);
      procedure _work_doModeDelim(var _Data: TData; Index: Word);
      procedure _work_doModeDisabled(var _Data: TData; Index: Word);
      procedure _work_doModeLen(var _Data: TData; Index: Word);
      procedure _work_doModeSimple(var _Data: TData; Index: Word);
      procedure _work_doModeTransp(var _Data: TData; Index: Word);
      procedure _work_doReset(var _Data: TData; Index: Word);
      procedure _work_doStore(var _Data: TData; Index: Word);
      
      procedure _var_CurrentData(var _Data: TData; Index: Word);
  end;

implementation

const
  MODE_SIMPLE = 0;
  MODE_DELIM = 1;
  MODE_LENGTH = 2;
  MODE_TRANSP = 3;
  MODE_DISABLED = 4;



constructor ThiDataAccumulator.Create;
begin
  //FPosInBuffer := 1;
end;

procedure ThiDataAccumulator.SetBufferSize(Value: Integer);
begin
  SetLength(FBuffer, Value);
  if FCurrentLength > Value then FCurrentLength := Value;
end;

procedure ThiDataAccumulator._work_doFinish;
begin
  if FCurrentLength > 0 then _hi_OnEvent(_event_onData, Copy(FBuffer, 1, FCurrentLength));
  _hi_CreateEvent(_Data, @_event_onFinish);
end;

procedure ThiDataAccumulator._work_doModeDelim;
begin
  FTotalLength := 0;
  _prop_Mode := MODE_DELIM;
end;

procedure ThiDataAccumulator._work_doModeDisabled;
begin
  _prop_Mode := MODE_DISABLED;
end;

procedure ThiDataAccumulator._work_doModeLen;
begin
  _prop_Mode := MODE_LENGTH;
  FTotalLength := 0;
end;

procedure ThiDataAccumulator._work_doModeTransp;
begin
  _prop_Mode := MODE_TRANSP;
end;

procedure ThiDataAccumulator._work_doModeSimple;
begin
  _prop_Mode := MODE_SIMPLE;
end;

procedure ThiDataAccumulator._work_doReset;
begin
  FCurrentLength := 0;
  FTotalLength := 0;
end;

//      SearchBuf   Buffer
function PosBuf(Buffer: Pointer; BufLen: Integer; SearchBuf: Pointer; SBufLen: Integer): Integer;
var B, Tmp: PChar;
begin
  Result := -1;
  if (BufLen < 1) or (SBufLen < 1) or (BufLen < SBufLen) then exit;
  B := Buffer;
  Tmp := B + BufLen - SBufLen;
  while B <= Tmp do
  begin
    if B^ = PChar(SearchBuf)^ then
    begin
      if CompareMem(B, SearchBuf, SBufLen) then // 
      begin
        Result := B - Buffer;
        break;
      end;
    end;
    Inc(B);
  end;
end;

//Simple,Delimiter,Length,Transparent,Disabled
procedure ThiDataAccumulator._work_doStore;
label
  next_mode;
var
  S, D, Rslt: string;
  LS, LB, L, I, C, K: Integer;
  
  procedure ShiftBuffer(const FromPos: Integer);
  begin
    if FromPos > FCurrentLength then
    begin
      FCurrentLength := 0;
      Exit;
    end;
    FCurrentLength := FCurrentLength - FromPos + 1;
    Move(FBuffer[FromPos], FBuffer[1], FCurrentLength);
  end;
begin

  S := ReadString(_Data, _data_Data, '');
  LB := Length(FBuffer);

next_mode: //      :   -  
  if _prop_Mode = MODE_DISABLED then Exit;
  
  LS := Length(S);
  //if (LS = 0) or (LB = 0) then Exit;
  
  case _prop_Mode of
    MODE_SIMPLE:
      begin
        I := 0; //      
        while I < LS do
        begin
          K := LB - FCurrentLength; //     
          C := LS - I; //    
          if C > K then C := K; //     
          Move(S[I+1], FBuffer[FCurrentLength+1], C);
          Inc(I, C);
          Inc(FCurrentLength, C);
          if FCurrentLength = LB then //   -     
          begin
            Rslt := Copy(FBuffer, 1, LB);
            _hi_OnEvent(_event_onData, Rslt);
            FCurrentLength := 0;
            if _prop_Mode <> MODE_SIMPLE then
            begin
              S := Copy(S, I+1, LS-I);
              goto next_mode;
            end;
          end;
        end;
      end;
      
    MODE_DELIM:
      begin
        D := ReadString(_Data, _data_Delimiter, _prop_Delimiter);
        L := Length(D);
        if L = 0 then Exit; //?
        
        //       
        K := PosBuf(@FBuffer[1], FCurrentLength, @D[1], L); //   (0-based)
        if K > -1 then //  
        begin
          Rslt := Copy(FBuffer, 1, K); //     
          ShiftBuffer(K+1+L); //         
          FTotalLength := 0;
          if K > 0 then _hi_OnEvent(_event_onData, Rslt); //  K = 0 Length(Rslt)=0,   
          _hi_OnEvent(_event_onFinish);
          goto next_mode; //   ,          ,   
        end;

        
        I := 0; //      
        while I < LS do
        begin
          K := LB - FCurrentLength; //     
          C := LS - I; //    
          if C > K then C := K; //     
          Move(S[I+1], FBuffer[FCurrentLength+1], C);
          Inc(I, C);
          Inc(FCurrentLength, C);
          Inc(FTotalLength, C);
          
          {
            //TODO:      
            K := PosBuf(@FBuffer[FCurrentLength-L+2], C+L-1, @D[1], L);
            Inc(FCurrentLength, C);
            Inc(FTotalLength, C);
          }
          
          K := PosBuf(@FBuffer[1], FCurrentLength, @D[1], L); //   (0-based)
          if K > -1 then //  
          begin
            Rslt := Copy(FBuffer, 1, K); //     
            ShiftBuffer(K+1+L); //         
            FTotalLength := 0;
            if K > 0 then _hi_OnEvent(_event_onData, Rslt); //  K = 0 Length(Rslt)=0,   
            _hi_OnEvent(_event_onFinish);
          end
          else
          begin
            if FTotalLength >= Cardinal(_prop_MaxLength) then //      
            begin
              FTotalLength := 0;
              _hi_OnEvent(_event_onError);
              Exit;
            end;
            if FCurrentLength = LB then //   -       L-1,  .  
            begin
              K := L - 1; // ,           
              C := LB - K;
              Rslt := Copy(FBuffer, 1, C); //   
              ShiftBuffer(C+1); //    
              _hi_OnEvent(_event_onData, Rslt);
            end;
          end;
          if _prop_Mode <> MODE_DELIM then
          begin
            S := Copy(S, I+1, LS-I);
            goto next_mode;
          end;
        end;      
      end;
      
    MODE_LENGTH:
      begin
        L := ReadInteger(_Data, _data_Length, _prop_Length);
        if L = 0 then Exit; //?
        
        //       
        K := Cardinal(L) - FTotalLength; //       
        if Cardinal(FCurrentLength) >= Cardinal(K) then //      
        begin
          Rslt := Copy(FBuffer, 1, K); //  
          ShiftBuffer(K+1); //       
          FTotalLength := 0;
          _hi_OnEvent(_event_onData, Rslt);
          _hi_OnEvent(_event_onFinish);
          goto next_mode; //   
        end;
        
        I := 0; //      
        while I < LS do
        begin
          K := LB - FCurrentLength; //     
          C := LS - I; //    
          if C > K then C := K; //     
          Move(S[I+1], FBuffer[FCurrentLength+1], C);
          Inc(I, C);
          Inc(FCurrentLength, C);
          
          K := Cardinal(L) - FTotalLength; //       
          if Cardinal(FCurrentLength) >= Cardinal(K) then //      
          begin
            Rslt := Copy(FBuffer, 1, K); //    
            ShiftBuffer(K+1); //       
            FTotalLength := 0;
            _hi_OnEvent(_event_onData, Rslt);
            _hi_OnEvent(_event_onFinish);
          end
          else
          begin
            if FCurrentLength = LB then //   -     
            begin
              Rslt := Copy(FBuffer, 1, LB);
              FCurrentLength := 0;
              Inc(FTotalLength, LB);
              _hi_OnEvent(_event_onData, Rslt);
            end;
          end;
          if _prop_Mode <> MODE_LENGTH then
          begin
            S := Copy(S, I+1, LS-I);
            goto next_mode;
          end;
        end;
      end;
      
    MODE_TRANSP:
      _hi_CreateEvent(_Data, @_event_onData, S);
  end;
end;

procedure ThiDataAccumulator._var_CurrentData;
begin
  dtString(_Data, Copy(FBuffer, 1, FCurrentLength));
end;



{.$WARNINGS ON}

end.
